using System;
using System.Collections.Generic;
using System.Text;

namespace SpringPCSCAPI
{
  public class APDUCmd
  {
    // sending variables
    private byte _CLA;
    private byte _INS;
    private byte _P1;
    private byte _P2;
    private byte _LC;
    private byte[] _DataIn;
    private byte _LE;

    private byte[] _APDUArray;
    private string _APDUStr;
    private string _DataInStr;


    // construct a sending APDU
    public APDUCmd(byte Class, byte Instruction, byte Parameter1, byte Parameter2) 
    {
      _CLA = Class;
      _INS = Instruction;
      _P1 = Parameter1;
      _P2 = Parameter2;
      _APDUArray = new byte []{Class, Instruction, Parameter1, Parameter2};
      _APDUStr = createStrFromByteArray(_APDUArray);
    }

    public APDUCmd( byte Class, byte Instruction, byte Parameter1, byte Parameter2, 
      byte[] CommandData)
    {
      _CLA = Class;
      _INS = Instruction;
      _P1 = Parameter1;
      _P2 = Parameter2;
      _LC = (byte) CommandData.Length;
      _DataIn = CommandData;
      _DataInStr = createStrFromByteArray(_DataIn);
      _APDUArray = new byte[5 + CommandData.Length];
      _APDUArray[0] = Class;
      _APDUArray[1] = Instruction;
      _APDUArray[2] = Parameter1;
      _APDUArray[3] = Parameter2;
      _APDUArray[4] = (byte) CommandData.Length;
      for (int cpt = 0; cpt < CommandData.Length; cpt++)
        _APDUArray[cpt + 5] = CommandData[cpt];
      _APDUStr = createStrFromByteArray(_APDUArray);
    }

    public APDUCmd(byte Class, byte Instruction, byte Parameter1, byte Parameter2,
      string CommandData)
    {
      _CLA = Class;
      _INS = Instruction;
      _P1 = Parameter1;
      _P2 = Parameter2;
      _LC = (byte)CommandData.Length;
      _DataIn = createByteArrayFromStr(CommandData);
      _DataInStr = createStrFromByteArray(_DataIn);
      _APDUArray = new byte[5 + _DataIn.Length];
      _APDUArray[0] = Class;
      _APDUArray[1] = Instruction;
      _APDUArray[2] = Parameter1;
      _APDUArray[3] = Parameter2;
      _APDUArray[4] = (byte)_DataIn.Length;
      for (int cpt = 0; cpt < _DataIn.Length; cpt++)
        _APDUArray[cpt + 5] = _DataIn[cpt];
      _APDUStr = createStrFromByteArray(_APDUArray);
    }

    public APDUCmd(byte Class, byte Instruction, byte Parameter1, 
      byte Parameter2, byte LengthExpected)
    {
      _CLA = Class;
      _INS = Instruction;
      _P1 = Parameter1;
      _P2 = Parameter2;
      _LE = LengthExpected;
      _APDUArray =  new byte[] { Class, Instruction, Parameter1, 
        Parameter2, LengthExpected};
      _APDUStr = createStrFromByteArray(_APDUArray);
    }

    public APDUCmd( byte Class, byte Instruction, byte Parameter1, 
      byte Parameter2, byte[] CommandData, byte LengthExpected)
    {
      _CLA = Class;
      _INS = Instruction;
      _P1 = Parameter1;
      _P2 = Parameter2;
      _LC = (byte)CommandData.Length;
      _DataIn = CommandData;
      _DataInStr = createStrFromByteArray(_DataIn);
      _LE = LengthExpected;
      _APDUArray = new byte[6 + CommandData.Length];
      _APDUArray[0] = Class;
      _APDUArray[1] = Instruction;
      _APDUArray[2] = Parameter1;
      _APDUArray[3] = Parameter2;
      _APDUArray[4] = (byte) CommandData.Length;
      for (int cpt = 0; cpt < CommandData.Length; cpt++)
        _APDUArray[cpt + 5] = CommandData[cpt];
      _APDUArray[CommandData.Length + 5] = LengthExpected;
      _APDUStr = createStrFromByteArray(_APDUArray);
    }

    public APDUCmd(byte Class, byte Instruction, byte Parameter1,
      byte Parameter2, string CommandData, byte LengthExpected)
    {
      _CLA = Class;
      _INS = Instruction;
      _P1 = Parameter1;
      _P2 = Parameter2;
      _LC = (byte)CommandData.Length;
      _DataIn = createByteArrayFromStr(CommandData);
      _DataInStr = createStrFromByteArray(_DataIn);
      _LE = LengthExpected;
      _APDUArray = new byte[6 + CommandData.Length];
      _APDUArray[0] = Class;
      _APDUArray[1] = Instruction;
      _APDUArray[2] = Parameter1;
      _APDUArray[3] = Parameter2;
      _APDUArray[4] = (byte)_DataIn.Length;
      for (int cpt = 0; cpt < _DataIn.Length; cpt++)
        _APDUArray[cpt + 5] = _DataIn[cpt];
      _APDUArray[CommandData.Length + 5] = LengthExpected;
      _APDUStr = createStrFromByteArray(_APDUArray);
    }

    // accept 0 arg
    public APDUCmd()
    {
      
    }

    // string argument, a raw Apdu 
    // in this case, LC is given when needed
    public APDUCmd(string RawApdu)
    {
      byte[] FullApduArray = createByteArrayFromStr(RawApdu);
      _APDUArray = FullApduArray;
      _APDUStr = RawApdu;

      if (FullApduArray.Length >= 4)
      {
        _CLA = FullApduArray[0];
        _INS = FullApduArray[1];
        _P1 = FullApduArray[2];
        _P2 = FullApduArray[3];
      }

      if (FullApduArray.Length == 5)
        _LE = FullApduArray[4];
      
      if(FullApduArray.Length >5)
      {
        // here we have to determine if LE is present or not
        // LC will give us DataIn size
        _LC = FullApduArray[4];
        _DataIn = new byte[_LC];
      

        if (FullApduArray.Length > _LC + 5)
        {
          //last byte is LE
          if (FullApduArray.Length - 2 >= 5)
          {
            for (int cpt = 5; cpt < FullApduArray.Length - 1; cpt++)
              _DataIn[cpt - 5] = FullApduArray[cpt];
          }
          _DataInStr = createStrFromByteArray(_DataIn);
          _LE = FullApduArray[FullApduArray.Length - 1];
        }
        else
        {
          for (int cpt = 5; cpt < FullApduArray.Length ; cpt++)
            _DataIn[cpt-5] = FullApduArray[cpt];
          _DataInStr = createStrFromByteArray(_DataIn);
        }
        
      }

    }


    private string createStrFromByteArray(byte[] byteArray) 
    {
      string Str = "";
      if (byteArray.Length <= 0)
        return Str;
      for (int cpt = 0; cpt < byteArray.Length - 1; cpt++)
        Str += String.Format("{0:X02}", byteArray[cpt]) + " ";
      Str += String.Format("{0:X02}", byteArray[byteArray.Length - 1]);
      return Str;
    }

    private byte[] createByteArrayFromStr(string StrIn)
    {
      string Str = StrIn.Replace("0x", "");

      if (!Str.Contains(" "))
      {
        string nStr = "";
        for (int cpt = 0; cpt < Str.Length; cpt = cpt + 2)
        {
          nStr = nStr + Str.Substring(cpt, 2) + " ";
        }
        Str = nStr;
      }
   
      // command stored into byte array
      char[] delimiterByte = new char[1];
      delimiterByte[0] = ' ';
      string properStr = Str;
      // removing last " "
      if (Str.EndsWith(" "))
      {
        properStr = Str.Substring(0, Str.Length - 1);
      }

      string[] Strbuff = properStr.Split(delimiterByte);
      byte[] StrBuffer = new byte[Strbuff.Length];
      for (int i = 0; i < Strbuff.Length; i++)
      {
        // converting byte string to byte  
        StrBuffer[i] = System.Convert.ToByte(int.Parse(Strbuff[i],
          System.Globalization.NumberStyles.HexNumber));
      }
      return StrBuffer;
    }


    public override string ToString() 
    {
      return _APDUStr;
    }

    public byte[] ToByteArray()
    {
      return _APDUArray;
    }

    public byte CLA 
    {
      get { return _CLA; }
      //set { _CLA = value; }
    }

    public byte INS
    {
      get { return _INS; }
      //set { _INS = value; }
    }

    public byte P1
    {
      get { return _P1; }
      //set { _P1 = value; }
    }
    public byte P2
    {
      get { return _P2; }
      //set { _P2 = value;}
    }
    public byte LC
    {
      get { return _LC; }
      //set { _LC = value;}
    }

    public byte[] DataIn
    {
      get { return _DataIn; }
      //set { _DataIn = value;}
    }

    public byte LE
    {
      get { return _LE; }
      //set { _LE = value;}
    }


    public byte[] APDUArray
    {
      get { return _APDUArray; }
      //set { _APDUArray = value;}
    }

    public string APDUStr
    {
      get { return _APDUStr; }
      //set { _APDUStr = value;}
    }

    public string DataInStr
    {
      get { return _DataInStr; }
      //set { _DataInStr = value;}
    }






   }
}
