using System;
using System.IO;
using System.Collections.ObjectModel;
using System.Collections;
using System.Runtime.InteropServices;
using System.Security.Cryptography;



namespace SpringPCSCAPI
{
  public abstract partial class SPCSC
  {
 #region constants
    //public const int SCARD_S_SUCCESS = 0;
    public const int SCARD_PROTOCOL_T0 = 1;
    public const int SCARD_PROTOCOL_T1 = 2;
    public const int TRANSMIT_ERR = -2;
    public const int AUTH_ERR = -3;
    public const int SPCSC_ERR = -1;
    public const int SPCSC_SUCCESS = 0;
    public const byte KEY_TYPE1 = 0x00;
    public const byte KEY_TYPE2 = 0x40;
    public const int MAX_RECV_LEN = 260;
    private static string ReaderName ="";
    private static int nCard;
    private static int nContext;
    private static bool loaded = false;
    public const int SCARD_LEAVE_CARD = 0; // Don't do anything special on close
    public const int SCARD_RESET_CARD = 1; // Reset the card on close
    public const int SCARD_UNPOWER_CARD = 2; // Power down the card on close
    public const int SCARD_EJECT_CARD = 3; // Eject the card on close

    // keys to be loaded onto reader's volatile memory
    // only 8 keys (4 type A and 4 type B)
    private static byte[][] keys  = new byte[8][];
    private static byte[] keyNumbers = new byte[8];
    private static int nbKeys = 0;

    /*
     * Structure used in GetStatusChange
     */

    [StructLayout(LayoutKind.Sequential, CharSet = CharSet.Auto)]
    public struct SCARD_READERSTATE
    {
      internal string szReader;
      internal IntPtr pvUserData;
      internal uint dwCurrentState;
      internal uint dwEventState;
      internal uint cbAtr;
      [MarshalAs(UnmanagedType.ByValArray, SizeConst = 0x24,
      ArraySubType = UnmanagedType.U1)]
      internal byte[] rgbAtr;
    }

   
    
#endregion


#region reader connection

    [DllImport("WinScard.dll", EntryPoint = "SCardEstablishContext")]
    public static extern uint EstablishContext(uint dwScope, int nNotUsed1,
      int nNotUsed2, ref int phContext);
    [DllImport("WinScard.dll", EntryPoint = "SCardConnect")]
    public static extern uint Connect(int hContext, string cReaderName,
      uint dwShareMode, uint dwPrefProtocol, ref int phCard,
      ref int ActiveProtocol);
    [DllImport("winscard.dll", EntryPoint = "SCardListReaders",
      SetLastError = true, CharSet = CharSet.Unicode)]
    internal static extern int ListReaders(int context, string groups,
      string readers, ref int size);
    [DllImport("winscard.dll", EntryPoint = "SCardGetStatusChange", CharSet = CharSet.Unicode)]
    public static extern uint GetStatusChange(int hContext, int dwTimeout,
      [In, Out, MarshalAs(UnmanagedType.LPArray, SizeParamIndex = 3)] 
      SCARD_READERSTATE[] rgReaderState, uint cReaders);
    [DllImport("winscard.dll", EntryPoint = "SCardReconnect")]
    public static extern uint Reconnect(int hCard, uint dwShareMode,
      uint dwPrefProtocol, uint swInit, ref int ActiveProtocol);


    public static string[] getListReaders() 
    {
      //First step in using smart cards is CSardEstablishContext()
      uint hContext = 2;		//system
      int nNotUsed1 = 0;
      int nNotUsed2 = 0;
      nContext = 0;		//handle to context for future calls to use

      uint RetEstablish = EstablishContext(hContext, nNotUsed1, nNotUsed2,
        ref nContext);

      int size = 2048;
      int nbReaders = 0; 
      Collection<string> readers = new Collection<string>();
      if (ListReaders(nContext, null, null, ref size) == SCARD_S_SUCCESS)
      {
        // Allocate a string of the proper size.
        string names = new string(' ', size);
        if (ListReaders(nContext, null, names, ref size) ==
          SCARD_S_SUCCESS)
        {
          // The 'names' string will contain a multi-string of the,
          // reader names i.e. they are separated by 0x00 characters.
          string name = string.Empty;

          for (int i = 0; i < names.Length; i++)
          {
            if (names[i] == 0x00)
            {
              if (name.Length > 0)
              {
                //
                // We have the reader name.
                //
                readers.Add(name);
                name = string.Empty;
                nbReaders++;
              }
            }
            else
            {
              // Append found character.
              name = name + new string(names[i], 1);
            }
          }
        }
      }
      
      string[] cReaders = new string[nbReaders];
      readers.CopyTo(cReaders, 0);
      
      return cReaders;
    }

    // simple function to set reader name, important parameter for ConnectCard
    public static void SetReader(string Name)
    {
      ReaderName = Name;
      nCard = 0;
    }

    public static int ConnectCard()
    {
      //First step in using smart cards is CSardEstablishContext()
      uint hContext = 2;		//system
      int nNotUsed1 = 0;
      int nNotUsed2 = 0;
      int nContext = 0;		//handle to context for future calls to use

      uint RetEstablish = EstablishContext(hContext, nNotUsed1, nNotUsed2,
        ref nContext);

      uint nShareMode = 1;			//exclusive
      int ActiveProtocol = SCARD_PROTOCOL_T0 | SCARD_PROTOCOL_T1 ;
      uint nPrefProtocol = (uint)ActiveProtocol;
      uint RetConnect = Connect(nContext, ReaderName,
        nShareMode, nPrefProtocol, ref nCard, ref ActiveProtocol);
      if (RetConnect != SCARD_S_SUCCESS)
        return SPCSC_ERR;
      else
        return 0;
 
    }

    public static int Reconnect()
    {
      uint nShareMode = 1;
      uint swInit = SCARD_RESET_CARD; 

      int ActiveProtocol = SCARD_PROTOCOL_T0 | SCARD_PROTOCOL_T1;
      uint nPrefProtocol = (uint)ActiveProtocol;

      uint RetConnect = Reconnect(nCard, nShareMode, nPrefProtocol,
          swInit, ref ActiveProtocol);
      if (RetConnect != SCARD_S_SUCCESS)
        return SPCSC_ERR;
      else
        return 0;

    }

    public static int SetKey(byte[] key, byte keyNumber)
    {
      // size expected is 6
      if ((key.Length != 6) || (nbKeys >7))
        return SPCSC_ERR;

      keys[nbKeys] = new byte[6];
      Array.Copy(key,keys[nbKeys],6);
      keyNumbers[nbKeys] = keyNumber;

      nbKeys++;

      return SPCSC_SUCCESS;
    }

    private static APDURsp LoadKey(byte[]key, byte keyNumber)
    {
      APDUCmd loadKeyApdu = new APDUCmd(0xFF, 0x82, 0x00, keyNumber, key);
      APDURsp answer = new APDURsp();
      Transmit(loadKeyApdu,ref answer);
      return (answer);
 
    }

    public static string StateToStr(uint State)
    {
      uint empty = State & SCARD_STATE_EMPTY;
      uint unaware = State & SCARD_STATE_UNAWARE;
      uint ignore = State & SCARD_STATE_IGNORE;
      uint changed = State & SCARD_STATE_CHANGED;
      uint unknown = State & SCARD_STATE_UNKNOWN;
      uint unavailable = State & SCARD_STATE_UNAVAILABLE;
      uint present = State & SCARD_STATE_PRESENT;
      uint atrMatch = State & SCARD_STATE_ATRMATCH;
      uint exclusive = State & SCARD_STATE_EXCLUSIVE;
      uint inuse = State & SCARD_STATE_INUSE;
      uint mute = State & SCARD_STATE_MUTE;

      if (mute != 0)
        return "SCARD_STATE_MUTE: Mute card in the reader. Try again";
      if (inuse != 0)
        return "SCARD_STATE_INUSE: There is a card in a reader, avalaible for use";
      if (exclusive != 0)
        return "SCARD_STATE_EXCLUSIVE: The card inserted is already in use in Exclusive mode";
      if (atrMatch != 0)
        return "SCARD_STATE_ATRMATCH: inserted card's ATR matches";
      if (present != 0)
        return "SCARD_STATE_PRESENT: Card present. Connect or Reconnect before use ";
      if (empty != 0)
        return "SCARD_STATE_EMPTY: No Card found";
      if (unavailable != 0)
        return "SCARD_STATE_UNAVAILABLE: Status unavailable";
      if (unknown != 0)
        return "SCARD_STATE_UNKNOWN: Reader unknown";
      if (changed != 0)
        return "SCARD_STATE_CHANGED: State has changed";
      if (ignore != 0)
        return "SCARD_STATE_IGNORE: Reader ignored";
      if (unaware != 0)
        return "SCARD_STATE_UNAWARE: Status wanted";
      return "Unknown state code ";
      
    }

    public static string CheckStatus()
    {
      string status = "";
      //checking only if a card is set
      if (ReaderName != "")
      {
        // Getting card status
        int TimeoutInf = 0; //infinite timeout
        uint numreaders = 1;
        SCARD_READERSTATE[] states = new SCARD_READERSTATE[1];
        states[0] = new SCARD_READERSTATE();
        states[0].szReader = ReaderName;
        states[0].pvUserData = IntPtr.Zero;
        states[0].dwCurrentState = 0;
        states[0].dwEventState = 0;
        states[0].cbAtr = 0;
        states[0].rgbAtr = null;
        uint retStatusChange = 0;

        retStatusChange = GetStatusChange(nContext, TimeoutInf,
          states, numreaders);

      

        if (retStatusChange != SCARD_S_SUCCESS)
        {
          status = "Error during GetStatusChange()" +
            retStatusChange.ToString();
        }
        else
        { // displaying current status
          status = StateToStr(states[0].dwEventState);
        }
      }
      return status;
    }



#endregion

#region card communication
    // from winscard : SCardTransmit : basic function to exchange data with card
    [DllImport("winscard.dll", EntryPoint = "SCardTransmit", SetLastError = true)]
    internal static extern uint TransmitCard(int hCard,
      IntPtr pioSendPci, byte[] pbSendBuffer,
      uint cbSendLength, IntPtr pioRecvPci, [In, Out] byte[] pbRecvBuffer,
      [In, Out] ref uint pcbRecvLength);

    // from winscard : SCardStatus : function used to retreive card's ATR
    [DllImport("winscard.dll", EntryPoint = "SCardStatus", SetLastError = true)]
    internal static extern uint StatusCard(int hCard, out IntPtr mszReaderNames,
      ref uint pcchReaderLen, out uint readerState, out uint protocol,
      [In,Out] byte[]ATRarray, ref uint ATRarryLength);


    [DllImport("WinScard.dll", EntryPoint = "SCardDisconnect")]
    public static extern uint DisconnectCard(int hCard, int Disposition );

    //IntPtr hCard, out IntPtr mszReaderNames, ref uint pcchReaderLen, out uint pdwState, 
    //out uint pdwProtocol, [In, Out] byte [] pbAtr, ref uint pbcAtrLen);


    /*
     * AtrLen=32;
			ReaderLen=SCARD_AUTOALLOCATE;

			if (SCARD_S_SUCCESS!=(rval=SCardStatus(hCard, out hReaders, ref ReaderLen , out CardState, out CardProtocol, Atr,  ref AtrLen)))
     * 
     * */
    public static bool Transmit(APDUCmd APDUIn, ref APDURsp APDUOut) 
    {
      byte[] buffer = new byte[MAX_RECV_LEN];
      uint bufLen = (uint) buffer.Length;
      IntPtr SCARD_IO_REQUEST = IntPtr.Zero;
      uint RetTransmit = TransmitCard(nCard, SCARD_IO_REQUEST, APDUIn.APDUArray, 
        (uint)APDUIn.APDUArray.Length, IntPtr.Zero, buffer,ref bufLen);
      byte[] resp = new byte[bufLen];
      for (int cpt = 0; cpt < bufLen; cpt++)
        resp[cpt] = buffer[cpt];
      APDUOut = new APDURsp(resp);
      if (RetTransmit != SCARD_S_SUCCESS)
        return false;
      return true;
    }

    public static string GetAtr()
    {
      string AtrStr = "";
      uint readerState=0 ;
     /* int protocol =0;
      byte[] buffer = new byte[MAX_RECV_LEN];*/
      uint protocol = 0;
      uint size = 32;
      uint ReaderLen = MAX_RECV_LEN;
      byte[] buffer = new byte[32];
      IntPtr readers = IntPtr.Zero;

      uint RetStatus = StatusCard(nCard,out readers,ref ReaderLen, out readerState,
        out protocol,buffer, ref size);
      
      AtrStr = RetStatus.ToString();
      
      /*if (RetStatus == SCARD_S_SUCCESS) 
      {
        for (int cpt = 0; cpt < size-1; cpt++)
          AtrStr += String.Format("{0:X02}", buffer[cpt]) + " ";
        AtrStr += String.Format("{0:X02}", buffer[size - 1]);
      }*/
      return AtrStr;
     }


     public static void Disconnect()
     {
       DisconnectCard(nCard, 0);
     }

    public static void Disconnect(int disposition)
    {
      DisconnectCard(nCard, disposition);
    }


     public static byte[] GetCardId()//reader's specific
     {
       APDUCmd GetIdAPDU = new APDUCmd(0xFF, 0xCA, 0xF0, 0x00, 0x00);
       DisconnectCard(nCard, 0);
       ConnectCard();
       APDURsp response = new APDURsp();
       Transmit(GetIdAPDU, ref response);
       return response.DataOut;
     }

     public static byte[] GetCardType()//reader's specific
     {
       APDUCmd GetTypeAPDU = new APDUCmd(0xFF, 0xCA, 0xF1, 0x00, 0x00);
       DisconnectCard(nCard, 0);
       ConnectCard();
       APDURsp response = new APDURsp();
       Transmit(GetTypeAPDU, ref response);
       return response.DataOut;
     }

    public static int GetHCard()
    {
      return nCard;
    }

#endregion

#region Mifare


    public static APDURsp ReadMif(byte sector, byte keyNumber) 
    {
      // load the keys if this is first connection 
      if (!loaded)
      {
        loaded = true;
        APDURsp retValue;
        for (int i = 0; i < nbKeys; i++)
        {
          retValue = LoadKey(keys[i], keyNumbers[i]);
          if ((retValue.SW1 != 0x90) || (retValue.SW2 != 0x00))
            return retValue;
        }
      }
      
      // Read Mifare Classic Command not supported
      //APDUCmd ReadApdu = new APDUCmd(0xFF, 0xF8, 0x00, sector, 0x30);
      // Therefore, we use general authenticate then read binary to read 
      
      // General Authenaticate command :
      // we have to authenticate on the first block of the sector
      byte block = (byte)((int)sector * 4);
      byte length = 0x30;
      if ((int)sector >= 32)
      {
        length = 0xF0;
        block = (byte) ( 128 + 16 *((int)sector - 32));
      }

      APDUCmd AuthApdu = new APDUCmd(0xFF,0x86,0x00,0x00,
        new byte[]{0x01,0x00,block,0x00,keyNumber});
      APDURsp recvBuf = new APDURsp();
      Transmit(AuthApdu, ref recvBuf);

      // if authenticate succeeded, we read the sector
      if ((recvBuf.SW1 == 0x90) && (recvBuf.SW2 == 0x00))
     {
        APDUCmd ReadbApdu = new APDUCmd(0xFF, 0xB0, 0x00, block, length);
        bool retTransmit = Transmit(ReadbApdu, ref recvBuf);
      }

      return recvBuf;
    }

    public static APDURsp WriteMif(byte sector, byte[] dataToWrite,byte keyNumber)
    {
      // Mifare classic Write not supported
      // therefore, we perform general authenticate then update Binary
      // APDUCmd WriteApdu = new APDUCmd(0xFF, 0xF9, 0x00, sector, dataToWrite);
      
      // determine block
      byte block = (byte)((int)sector * 4);
      byte[] DataIn;
      int cpt;
    
      // according to sector, we change dataToWrite's size
      if ((int)sector < 32)
      {
        // length of a whole sector : 48 bytes
        DataIn = new byte[48];
        for (cpt = 0; cpt < 48; cpt++)
          DataIn[cpt] = dataToWrite[cpt];
      }
      else
      {
        DataIn = new byte[240];
        block = (byte)(128 + 16 * ((int)sector - 32));
        for (cpt = 0; cpt < 240; cpt++)
          DataIn[cpt] = dataToWrite[cpt];
      }


      APDUCmd AuthApdu = new APDUCmd(0xFF, 0x86, 0x00, 0x00,
        new byte[] { 0x01, 0x00, block, 0x00, keyNumber });
      APDURsp recvBuf = new APDURsp();
      Transmit(AuthApdu, ref recvBuf);
      APDUCmd UpdatebApdu = new APDUCmd(0xFF, 0xD6, 0x00, block, DataIn);

      if ((recvBuf.SW1 == 0x90) && (recvBuf.SW2 == 0x00))
      {
        
        bool retTransmit = Transmit(UpdatebApdu, ref recvBuf);
      }
      
      return recvBuf ;
    }

    public static APDURsp UpdateAccessBlock(byte sector, byte keyNumber,
      byte[] newKeyA, byte[] newKeyB, byte AC0, byte AC1, byte AC2, byte AC3)
    {
      // determine block
      byte block = (byte)((int)sector * 4 +3);

      // first we try to read access block
      APDUCmd AuthApdu = new APDUCmd(0xFF, 0x86, 0x00, 0x00,
       new byte[] { 0x01, 0x00, block, 0x00, keyNumber });
      APDURsp recvBuf = new APDURsp();
      Transmit(AuthApdu, ref recvBuf);


      // we format access conditions 

      BitArray AC0b = new BitArray(new byte[] { AC0 });
      BitArray AC1b = new BitArray(new byte[] { AC1 });
      BitArray AC2b = new BitArray(new byte[] { AC2 });
      BitArray AC3b = new BitArray(new byte[] { AC3 });

      BitArray Byte6Array = new BitArray(new bool[]
        {!(AC0b[2]),!(AC1b[2]),!(AC2b[2]),!(AC3b[2]),
          !(AC0b[1]),!(AC1b[1]),!(AC2b[1]),!(AC3b[1])});
      BitArray Byte7Array = new BitArray(new bool[] 
        {!(AC0b[0]),!(AC1b[0]),!(AC2b[0]),!(AC3b[0]),
          AC0b[2],AC1b[2],AC2b[2],AC3b[2]});
      BitArray Byte8Array = new BitArray(new bool[] 
        { AC0b[1], AC1b[1], AC2b[1], AC3b[1], AC0b[0], 
          AC1b[0], AC2b[0], AC3b[0] });

      byte[] condition0 = new byte[1];
      Byte6Array.CopyTo(condition0, 0);
      byte[] condition1 = new byte[1];
      Byte7Array.CopyTo(condition1, 0);
      byte[] condition2 = new byte[1];
      Byte8Array.CopyTo(condition2, 0);

      byte conditon3 = new byte();
      

      //if read succeed, we update with our key
      if ((recvBuf.SW1 == 0x90) && (recvBuf.SW2 == 0x00))
      {
        APDUCmd UpdatebApdu = new APDUCmd(0xFF, 0xD6, 0x00, block,
            new byte[] { newKeyA[0], newKeyA[1], newKeyA[2], newKeyA[3], newKeyA[4], 
          newKeyA[5],condition0[0],condition1[0],condition2[0],conditon3, 
              newKeyB[0],newKeyB[1], newKeyB[2], newKeyB[3], newKeyB[4], newKeyB[5]});
        bool retTransmit = Transmit(UpdatebApdu, ref recvBuf);
      }

      return recvBuf;

    
    }

    


#endregion

#region Jaycos

    public static int AuthenticateJaycos()
    {
      // Phase 1 : get challenge 
      APDUCmd getChallengeApdu = new APDUCmd(0x00, 0x84, 0x00, 0x00, 0x08);
      APDURsp ChallengeApdu = new APDURsp();
      bool ret = Transmit(getChallengeApdu, ref ChallengeApdu);
      if (!ret)
        return TRANSMIT_ERR;

      // Phase 2 : encrypt challenge

      byte[] PasswordByteArray = new byte[]{0xFA, 0x1F, 0x8D, 0x09, 0x17, 
        0xF4, 0xD2, 0x2F, 0xC8, 0x71, 0xD2, 0x6A, 0x25, 0xD2, 0xD0, 0xFF};

      TripleDES alg = TripleDES.Create();
      alg.Key = PasswordByteArray;
      alg.Mode = CipherMode.ECB;

      // Now create a crypto stream through which we are going
      // to be pumping data. 
      // Our fileOut is going to be receiving the encrypted bytes. 
      MemoryStream ms = new MemoryStream();
      CryptoStream cs = new CryptoStream(ms,
            alg.CreateEncryptor(), CryptoStreamMode.Write);

      cs.Write(ChallengeApdu.DataOut, 0, ChallengeApdu.DataOut.Length);
      cs.Close();
      ms.Close();
      byte[] encryptedData = ms.ToArray();

      byte[] encryptedChallenge = new byte[] {
      encryptedData[0],encryptedData[1],encryptedData[2],encryptedData[3],
      encryptedData[4],encryptedData[5],encryptedData[6],encryptedData[7]};

      // Phase 3 : Sending encrypted challenge
      APDUCmd authenticateApdu = new APDUCmd(
        0x00, 0x84, 0x00, 0x00, encryptedChallenge);
      APDURsp authenticateSuccess = new APDURsp();
      ret = Transmit(authenticateApdu, ref authenticateSuccess);
      if (!ret)
        return TRANSMIT_ERR;
      if ((authenticateSuccess.SW1 == 0x90) && (authenticateSuccess.SW2 == 0x00))
        return  SPCSC_SUCCESS;
      else
        return AUTH_ERR;
    }

    public static byte[] ReadFileJaycos(byte[] filePath, int length)
    {
      byte[] fileData = new byte[length];
      int cpt=length;
      byte[] temp;

      // select file first
      APDUCmd selectFile = new APDUCmd(0x00,0xA4,0x08,0x00,
       filePath);
      APDURsp resp = new APDURsp();
      Transmit(selectFile,ref resp);
      //get response
      APDUCmd getResp = new APDUCmd(0x00, 0xC0, 0x00, 0x00, resp.SW2);
      Transmit(getResp,ref resp);

      bool loopAgain = true;
      byte start1 = 0x00;
      byte start2 = 0x00;
      int i;
      byte startIndex = 0;

      // read 
      do
      {
        if (cpt < 143)
        {
          temp = ReadIntoSelectedFileJaycos(start1, start2, (byte)cpt);
          for (i = startIndex;
            i < startIndex + cpt; i++)
            fileData[i] = temp[i - startIndex];
          loopAgain = false;
        }
        else 
        {
          temp = ReadIntoSelectedFileJaycos(start1, start2, 0x8E); //8E(hex)=142(dec) 
          for (i = startIndex;
            i <startIndex + 142; i++)
            fileData[i] = temp[(i - startIndex)];
          cpt -= 142;
          // add 142 to start index
          byte[] newIndexes = System.BitConverter.GetBytes(
            startIndex + 142);
          startIndex += 142;
          start1 = newIndexes[1];
          start2 = newIndexes[0];
        }

      } while (loopAgain);
      return (fileData);
    }

    public static byte[] ReadIntoSelectedFileJaycos(byte startIndex1,
      byte startIndex2, byte length)
    {
      APDURsp respFic = new APDURsp();
      APDUCmd read = new APDUCmd(0x00, 0xB0, startIndex1, startIndex2, length);
      Transmit(read, ref respFic);
      return respFic.DataOut;
    }

    public static APDUCmd WriteFileJaycos(byte[] filePath, byte[] DataToWrite)
    {
      int cpt = DataToWrite.Length;
      byte[] temp = new byte[142];

      // select file first
      APDUCmd selectFile = new APDUCmd(0x00, 0xA4, 0x08, 0x00,
       filePath);
      APDURsp resp = new APDURsp();
      Transmit(selectFile, ref resp);
      //get response
      APDUCmd getResp = new APDUCmd(0x00, 0xC0, 0x00, 0x00, resp.SW2);
      Transmit(getResp, ref resp);

      bool loopAgain = true;
      APDUCmd success;
      byte start1 = 0x00;
      byte start2 = 0x00;
      int i;
      byte startIndex = 0;

      // write
      do
      {
        if (cpt < 143)
        {
          // create temporary byte array to write
          byte[] temp2 = new byte[cpt];
          for (i = startIndex;
            i < startIndex + cpt; i++)
          temp2[i - startIndex] = DataToWrite[i];
          success = WriteIntoSelectedFileJaycos(start1, start2, temp2);
          loopAgain = false;
        }
        else
        {

          for (i = startIndex;
            i < startIndex + 142; i++)
            temp[(i - startIndex)] =DataToWrite[i];
          success = WriteIntoSelectedFileJaycos(start1, start2, temp); 
          cpt -= 142;
          // add 142 to start index
          byte[] newIndexes = System.BitConverter.GetBytes(
            startIndex + 142);
          startIndex += 142;
          start1 = newIndexes[1];
          start2 = newIndexes[0];
        }

      } while (loopAgain);
      return success;
    }

    public static APDUCmd WriteIntoSelectedFileJaycos(byte startIndex1,
      byte startIndex2, byte[] data)
    {
      APDURsp respFic = new APDURsp();
      APDUCmd write = new APDUCmd(0x00, 0xD6, startIndex1, startIndex2, data);
      Transmit(write, ref respFic);
      return write;
    }

    public static byte[] GetUserDataJaycos()
    {
      byte[] userData = new byte[128+340];
      int cpt;
      byte[] temp;
      
      // select user identity file first. Size : 128
      temp = ReadFileJaycos(new byte[] { 0x3F, 0x00, 0x91, 0x01, 0x00, 0x01 }, 128);
      for (cpt = 0; cpt < 128; cpt++)
        userData[cpt] = temp[cpt];
      

      // select fingerprints file. size : 340
      temp = ReadFileJaycos( new byte[] { 0x3F, 0x00, 0x91, 0x01, 0x00, 0x02 },340);
      for (cpt = 128; cpt < 128 + 340; cpt++)
        userData[cpt] = temp[cpt-128];


      // return all data fetched
      return (temp);
    }


    public static APDURsp ClearFileJaycos(byte[] filePath)
    { 
      // two steps :
      //    select file
      //    erase file content

      APDUCmd selectFile = new APDUCmd(0x00, 0xA4, 0x08, 0x00,
       filePath);
      APDURsp resp = new APDURsp();
      Transmit(selectFile, ref resp);
      //get response
      APDUCmd getResp = new APDUCmd(0x00, 0xC0, 0x00, 0x00, resp.SW2);
      Transmit(getResp, ref resp);

      if((resp.SW1 != 0x90) || (resp.SW2 != 0x00))
        return resp;

      APDUCmd ClearFile = new APDUCmd(0x00, 0x0E, 0x00, 0x00, 0x00);
      Transmit(ClearFile, ref resp);

      return resp;
    }

    public static APDUCmd FormatJaycosCard()
    {
      // we create our own file system on the card

      //first we authenticate
      AuthenticateJaycos();

      //create master file
      APDURsp res = new APDURsp();
      APDUCmd CreateMf = new APDUCmd("80 E0 3F 00 00");
      Transmit(CreateMf, ref res);

      //create directory file

      //create file
      return CreateMf;
    }

    

#endregion


#region error codes

    const uint SCARD_S_SUCCESS = 0x00000000;
    const uint SCARD_F_INTERNAL_ERROR = 0x80100001;
    const uint SCARD_E_CANCELLED = 0x80100002;
    const uint SCARD_E_INVALID_HANDLE = 0x80100003;
    const uint SCARD_E_INVALID_PARAMETER = 0x80100004;
    const uint SCARD_E_INVALID_TARGET = 0x80100005;
    const uint SCARD_E_NO_MEMORY = 0x80100006;
    const uint SCARD_F_WAITED_TOO_LONG = 0x80100007;
    const uint SCARD_E_INSUFFICIENT_BUFFER = 0x80100008;
    const uint SCARD_E_UNKNOWN_READER = 0x80100009;
    const uint SCARD_E_TIMEOUT = 0x8010000A;
    const uint SCARD_E_SHARING_VIOLATION = 0x8010000B;
    const uint SCARD_E_NO_SMARTCARD = 0x8010000C;
    const uint SCARD_E_UNKNOWN_CARD = 0x8010000D;
    const uint SCARD_E_CANT_DISPOSE = 0x8010000E;
    const uint SCARD_E_PROTO_MISMATCH = 0x8010000F;
    const uint SCARD_E_NOT_READY = 0x80100010;
    const uint SCARD_E_INVALID_VALUE = 0x80100011;
    const uint SCARD_E_SYSTEM_CANCELLED = 0x80100012;
    const uint SCARD_F_COMM_ERROR = 0x80100013;
    const uint SCARD_F_UNKNOWN_ERROR = 0x80100014;
    const uint SCARD_E_INVALID_ATR = 0x80100015;
    const uint SCARD_E_NOT_TRANSACTED = 0x80100016;
    const uint SCARD_E_READER_UNAVAILABLE = 0x80100017;
    const uint SCARD_P_SHUTDOWN = 0x80100018;
    const uint SCARD_E_PCI_TOO_SMALL = 0x80100019;
    const uint SCARD_E_READER_UNSUPPORTED = 0x8010001A;
    const uint SCARD_E_DUPLICATE_READER = 0x8010001B;
    const uint SCARD_E_CARD_UNSUPPORTED = 0x8010001C;
    const uint SCARD_E_NO_SERVICE = 0x8010001D;
    const uint SCARD_E_SERVICE_STOPPED = 0x8010001E;
    const uint SCARD_E_UNEXPECTED = 0x8010001F;
    const uint SCARD_E_ICC_INSTALLATION = 0x80100020;
    const uint SCARD_E_ICC_CREATEORDER = 0x80100021;
    const uint SCARD_E_UNSUPPORTED_FEATURE = 0x80100022;
    const uint SCARD_E_DIR_NOT_FOUND = 0x80100023;
    const uint SCARD_E_FILE_NOT_FOUND = 0x80100024;
    const uint SCARD_E_NO_DIR = 0x80100025;
    const uint SCARD_E_NO_FILE = 0x80100026;
    const uint SCARD_E_NO_ACCESS = 0x80100027;
    const uint SCARD_E_WRITE_TOO_MANY = 0x80100028;
    const uint SCARD_E_BAD_SEEK = 0x80100029;
    const uint SCARD_E_INVALID_CHV = 0x8010002A;
    const uint SCARD_E_UNKNOWN_RES_MNG = 0x8010002B;
    const uint SCARD_E_NO_SUCH_CERTIFICATE = 0x8010002C;
    const uint SCARD_E_CERTIFICATE_UNAVAILABLE = 0x8010002D;
    const uint SCARD_E_NO_READERS_AVAILABLE = 0x8010002E;
    const uint SCARD_E_COMM_DATA_LOST = 0x8010002F;
    const uint SCARD_E_NO_KEY_CONTAINER = 0x80100030;
    const uint SCARD_W_UNSUPPORTED_CARD = 0x80100065;
    const uint SCARD_W_UNRESPONSIVE_CARD = 0x80100066;
    const uint SCARD_W_UNPOWERED_CARD = 0x80100067;
    const uint SCARD_W_RESET_CARD = 0x80100068;
    const uint SCARD_W_REMOVED_CARD = 0x80100069;
    const uint SCARD_W_SECURITY_VIOLATION = 0x8010006A;
    const uint SCARD_W_WRONG_CHV = 0x8010006B;
    const uint SCARD_W_CHV_BLOCKED = 0x8010006C;
    const uint SCARD_W_EOF = 0x8010006D;
    const uint SCARD_W_CANCELLED_BY_USER = 0x8010006E;
    const uint SCARD_W_CARD_NOT_AUTHENTICATED = 0x8010006F;
    const uint SCARD_STATE_UNAWARE = 0x0000;
    const uint SCARD_STATE_IGNORE = 0x0001;
    const uint SCARD_STATE_CHANGED = 0x0002;
    const uint SCARD_STATE_UNKNOWN = 0x0004;
    const uint SCARD_STATE_UNAVAILABLE = 0x0008;
    const uint SCARD_STATE_EMPTY = 0x0010;
    const uint SCARD_STATE_PRESENT = 0x0020;
    const uint SCARD_STATE_ATRMATCH = 0x0040;
    const uint SCARD_STATE_EXCLUSIVE = 0x0080;
    const uint SCARD_STATE_INUSE = 0x0100;
    const uint SCARD_STATE_MUTE = 0x0200;


    public static string codeToStr(uint code)
    {
      switch (code)
      {
        case SCARD_S_SUCCESS: return "SCARD_S_SUCCESS";
        case SCARD_F_INTERNAL_ERROR: return "SCARD_F_INTERNAL_ERROR";
        case SCARD_E_CANCELLED: return "SCARD_E_CANCELLED";
        case SCARD_E_INVALID_HANDLE: return "SCARD_E_INVALID_HANDLE";
        case SCARD_E_INVALID_PARAMETER: return "SCARD_E_INVALID_PARAMETER";
        case SCARD_E_INVALID_TARGET: return "SCARD_E_INVALID_TARGET";
        case SCARD_E_NO_MEMORY: return "SCARD_E_NO_MEMORY";
        case SCARD_F_WAITED_TOO_LONG: return "SCARD_F_WAITED_TOO_LONG";
        case SCARD_E_INSUFFICIENT_BUFFER: return "SCARD_E_INSUFFICIENT_BUFFER";
        case SCARD_E_UNKNOWN_READER: return "SCARD_E_UNKNOWN_READER";
        case SCARD_E_TIMEOUT: return "SCARD_E_TIMEOUT";
        case SCARD_E_SHARING_VIOLATION: return "SCARD_E_SHARING_VIOLATION";
        case SCARD_E_NO_SMARTCARD: return "SCARD_E_NO_SMARTCARD";
        case SCARD_E_UNKNOWN_CARD: return "SCARD_E_UNKNOWN_CARD";
        case SCARD_E_CANT_DISPOSE: return "SCARD_E_CANT_DISPOSE";
        case SCARD_E_PROTO_MISMATCH: return "SCARD_E_PROTO_MISMATCH";
        case SCARD_E_NOT_READY: return "SCARD_E_NOT_READY";
        case SCARD_E_INVALID_VALUE: return "SCARD_E_INVALID_VALUE";
        case SCARD_E_SYSTEM_CANCELLED: return "SCARD_E_SYSTEM_CANCELLED";
        case SCARD_F_COMM_ERROR: return "SCARD_F_COMM_ERROR";
        case SCARD_F_UNKNOWN_ERROR: return "SCARD_F_UNKNOWN_ERROR";
        case SCARD_E_INVALID_ATR: return "SCARD_E_INVALID_ATR";
        case SCARD_E_NOT_TRANSACTED: return "SCARD_E_NOT_TRANSACTED";
        case SCARD_E_READER_UNAVAILABLE: return "SCARD_E_READER_UNAVAILABLE";
        case SCARD_P_SHUTDOWN: return "SCARD_P_SHUTDOWN";
        case SCARD_E_PCI_TOO_SMALL: return "SCARD_E_PCI_TOO_SMALL";
        case SCARD_E_READER_UNSUPPORTED: return "SCARD_E_READER_UNSUPPORTED";
        case SCARD_E_DUPLICATE_READER: return "SCARD_E_DUPLICATE_READER";
        case SCARD_E_CARD_UNSUPPORTED: return "SCARD_E_CARD_UNSUPPORTED";
        case SCARD_E_NO_SERVICE: return "SCARD_E_NO_SERVICE";
        case SCARD_E_SERVICE_STOPPED: return "SCARD_E_SERVICE_STOPPED";
        case SCARD_E_UNEXPECTED: return "SCARD_E_UNEXPECTED";
        case SCARD_E_ICC_INSTALLATION: return "SCARD_E_ICC_INSTALLATION";
        case SCARD_E_ICC_CREATEORDER: return "SCARD_E_ICC_CREATEORDER";
        case SCARD_E_UNSUPPORTED_FEATURE: return "SCARD_E_UNSUPPORTED_FEATURE";
        case SCARD_E_DIR_NOT_FOUND: return "SCARD_E_DIR_NOT_FOUND";
        case SCARD_E_FILE_NOT_FOUND: return "SCARD_E_FILE_NOT_FOUND";
        case SCARD_E_NO_DIR: return "SCARD_E_NO_DIR";
        case SCARD_E_NO_FILE: return "SCARD_E_NO_FILE";
        case SCARD_E_NO_ACCESS: return "SCARD_E_NO_ACCESS";
        case SCARD_E_WRITE_TOO_MANY: return "SCARD_E_WRITE_TOO_MANY";
        case SCARD_E_BAD_SEEK: return "SCARD_E_BAD_SEEK";
        case SCARD_E_INVALID_CHV: return "SCARD_E_INVALID_CHV";
        case SCARD_E_UNKNOWN_RES_MNG: return "SCARD_E_UNKNOWN_RES_MNG";
        case SCARD_E_NO_SUCH_CERTIFICATE: return "SCARD_E_NO_SUCH_CERTIFICATE";
        case SCARD_E_CERTIFICATE_UNAVAILABLE: return "SCARD_E_CERTIFICATE_UNAVAILABLE";
        case SCARD_E_NO_READERS_AVAILABLE: return "SCARD_E_NO_READERS_AVAILABLE";
        case SCARD_E_COMM_DATA_LOST: return "SCARD_E_COMM_DATA_LOST";
        case SCARD_E_NO_KEY_CONTAINER: return "SCARD_E_NO_KEY_CONTAINER";
        //case SCARD_E_SERVER_TOO_BUSY : return "SCARD_E_SERVER_TOO_BUSY";
        case SCARD_W_UNSUPPORTED_CARD: return "SCARD_W_UNSUPPORTED_CARD";
        case SCARD_W_UNRESPONSIVE_CARD: return "SCARD_W_UNRESPONSIVE_CARD";
        case SCARD_W_UNPOWERED_CARD: return "SCARD_W_UNPOWERED_CARD";
        case SCARD_W_RESET_CARD: return "SCARD_W_RESET_CARD";
        case SCARD_W_REMOVED_CARD: return "SCARD_W_REMOVED_CARD";
        case SCARD_W_SECURITY_VIOLATION: return "SCARD_W_SECURITY_VIOLATION";
        case SCARD_W_WRONG_CHV: return "SCARD_W_WRONG_CHV";
        case SCARD_W_CHV_BLOCKED: return "SCARD_W_CHV_BLOCKED";
        case SCARD_W_EOF: return "SCARD_W_EOF";
        case SCARD_W_CANCELLED_BY_USER: return "SCARD_W_CANCELLED_BY_USER";
        case SCARD_W_CARD_NOT_AUTHENTICATED: return "SCARD_W_CARD_NOT_AUTHENTICATED";
        default: return "";
      }
    }

    public static string StatusWordsToStr(byte SW1, byte SW2)
    {
      switch (SW1)
      {
        case 0x60: return "null";
        case 0x61: return "Still " + SW2.ToString() + " bytes available. Use GETRESPONSE (C0) to access this data";
        case 0x62:
          switch (SW2)
          {
            case 0x81: return "Warning : returned data may be corrupted";
            case 0x82: return "Warning : EoF has been reached before ending";
            case 0x83: return "Warning : selected file invalidated";
            case 0x84: return "Warning : bad file control information format";
            default: return "Warning : state unchanged";
          }
        case 0x63:
          if ((SW2 >= 0xC0) && (SW2 <= 0xCF))
            return "Warning : counter value is x";
          if (SW2 == 81)
            return "Warning : file filled up with last write";
          return "Warning : state unchanged";
        case 0x64: return "Error : state unchanged";
        case 0x65:
          switch (SW2)
          {
            case 0x01: return "Memory failure, problem in writing the EEPROM";
            case 0x81: return "Error : memory failure";
            default: return "Error : state changed";
          }
        case 0x66: return "Security error";
        case 0x67: return "Check error - incorrect byte length";
        case 0x68:
          switch (SW2)
          {
            case 0x81: return "Check error - logical channel not supported";
            case 0x82: return "Check error : secure messaging not supported";
            default: return "Check error : request function not supported";
          }
        case 0x69:
          switch (SW2)
          {
            case 0x81: return "Check error : command incompatible with file structure";
            case 0x82: return "Check error : security status not statisfied";
            case 0x83: return "Check error : authentication method blocked";
            case 0x84: return "Check error : referenced data invalidated";
            case 0x85: return "Check error : conditions of use not satisfied";
            case 0x86: return "Check error : command not allowed (no current EF)";
            case 0x87: return "Check error : Expected SM data objects missing";
            case 0x88: return "Check error : SM data objects incorrect ";
            default: return "Unknow command, most probably erroneous typing, protocol violation or incorrect format";
          }
        case 0x6A:
          switch (SW2)
          {
            case 0x00: return "Check error : P1 or P2 incorrect";
            case 0x80: return "Check error : parameters in data field incorrect";
            case 0x81: return "Check error : function not supported";
            case 0x82: return "Check error : file not found";
            case 0x83: return "Check error : record not found";
            case 0x84: return "Check error : insufficient memory space in this file";
            case 0x85: return "Check error : Lc inconsistant with TLV structure";
            case 0x86: return "Check error : inconsistant parameters P1-P2";
            case 0x87: return "Check error : P3 inconsistant with P1-P2";
            case 0x88: return "Check error : referenced data not found";
            default: return "Check error : wrong parameters";
          }

        case 0x6B: return "Check error : reference P1,P2 incorrect";
        case 0x6C: return "Lc length incorrect, correct length :" + SW2.ToString();
        case 0x6D: return "Ins invalid or unsupported";
        case 0x6E: return "Cla insupported";
        case 0x6F: return "Undiagnosted error";
        case 0x90:
          switch (SW2)
          {
            case 0x00: return "No Error";
            case 0x01: return "Failed to write EEPROM";
            case 0x10: return "Wrong PIN (1st try)";
            case 0x20: return "Wrong PIN(2nd try)";
            case 0x40: return "Wrong PIN (3rd try)";
            case 0x80: return "Wrong PIN, card blocked";
            default: return "Unknown error";
          }
        case 0x92:
          switch (SW2)
          {
            case 0x00: return "Reference executed ok";
            case 0x02: return "Failed to write EEPROM";
            default: return "Unknow error";
          }
        case 0x94: return "No EF selected";
        case 0x98:
          switch (SW2)
          {
            case 0x02: return "Invalid PIN";
            case 0x04: return "Wrong PIN presentation";
            case 0x06: return "PIN cancelled";
            case 0x08: return "PIN inactivated";
            case 0x10: return "Security condition unsatisfied";
            case 0x20: return "PIN inactive";
            default: return "Unknown error";
          }
        default: return "Unknown error";
      }



    }
   
#endregion

  }
}
